/*

   Fit function for 4 frequencies with 1st harmonic:
   Fit functions are implemented as external programms
   which will receive the following parameters:
   1. address of array (double) of x-values.
      (or shared memory ID in UNIX)
   2. address of array (double) of results.
      (or shared memory ID in UNIX)
   3. first array element
   4. last array element
   5. size of array in bytes (used for shmat() under UNIX !)
   6. 1st parameter (double)
   7. 2nd parameter (double)
      .
      .
      .
   26. last parameter (if used)
   On Amiga Computers the Address represents the
   real memory location in decimal format (%d)
   On UNIX Computers this parameter identifies
   a shared memory ID. See "examplefn.c"
   On Amiga Computers it is recommended to make
   the Fit function resident, since it will be
   called very often.

   This example contains the following parts:
   1. Commandline interface
   2. Help function (if no parameter was specified)
   3. (shared) Memory allocation
   4. a simple 'cos' function

*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#ifdef UNIX
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#endif

#define MAXPARM 30

#define sqr(x) ((x)*(x))
#define SIGN(a,b) ((b)<0 ? -fabs(a) : fabs(a))

double *p;
double *shmx, *shmy;
int   shmidx, shmidy;
int   first, last;
int   max;

int lorentz_flg;

/* ----------------------------------------------------------------
   perhaps you need a function to avoid overflow errors on division
   ---------------------------------------------------------------- */

double divide(a,b)
double a,b;
{
   if(b != 0.0) return((double)(a / b));
   return(1.0E10);
}

/* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   !!                 Here comes the function                     !!
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
double fn(x,a,b,c,a01,a11,w1,p1,d1,a02,a12,w2,p2,d2,
               a03,a13,w3,p3,d3,a04,a14,w4,p4,d4)
double    x,a,b,c,a01,a11,w1,p1,d1,a02,a12,w2,p2,d2,
               a03,a13,w3,p3,d3,a04,a14,w4,p4,d4;
{
double y, y1, y2, y3, y4;

   p1 = p1 * 0.01745329;
   p2 = p2 * 0.01745329;
   p3 = p3 * 0.01745329;
   p4 = p4 * 0.01745329;

   y1 = ((a01 * cos((w1 * x) + p1)) + (a11 * cos((2 * w1 * x) + p1)));
   y2 = ((a02 * cos((w2 * x) + p2)) + (a12 * cos((2 * w2 * x) + p2)));
   y3 = ((a03 * cos((w3 * x) + p3)) + (a13 * cos((2 * w3 * x) + p3)));
   y4 = ((a04 * cos((w4 * x) + p4)) + (a14 * cos((2 * w4 * x) + p4)));

   switch(lorentz_flg) {
      case 0: /* Gaussian damping */
         y1 = y1 * exp(-1.0 * (d1 * x) * (d1 * x));
         y2 = y2 * exp(-1.0 * (d2 * x) * (d2 * x));
         y3 = y3 * exp(-1.0 * (d3 * x) * (d3 * x));
         y4 = y4 * exp(-1.0 * (d4 * x) * (d4 * x));
         break;
      case 1:  /* Lorentz damping */
         y1 = y1 * exp(-1.0 * (d1 * x));
         y2 = y2 * exp(-1.0 * (d2 * x));
         y3 = y3 * exp(-1.0 * (d3 * x));
         y4 = y4 * exp(-1.0 * (d4 * x));
         break;
   }
   y = a + (b * x) + (c * x * x) + y1 + y2 + y3 + y4;

   return(y);
}

/* ----------------------------------------------------------------
   You should allways supply a help function, which will be called
   when no parameter is specified.
   ---------------------------------------------------------------- */

help()
{
printf("Fit function for 4 frequencies with 1st harmonic:\n");
printf("f(x) =  a + (b * x) + (c * x * x) +\n");
printf("        SUM(n=1..4) Damp(Dn,x) * [A0n * cos(wn * x + Pn) +\n");
printf("                                  A1n * cos(2 * Fn * x + Pn)]\n");
printf("Parameter:\n");
printf("p[0] = a ,   p[1] = b ,   p[2] = c \n");
printf("p[3]  = A01 , p[4]  = A11 , p[5]  = w1 , p[6]  = P1 , p[7]  = D1\n");
printf("p[8]  = A02 , p[9]  = A12 , p[10] = w2 , p[11] = P2 , p[12] = D2\n");
printf("p[13] = A03 , p[14] = A13 , p[15] = w3 , p[16] = P3 , p[17] = D3\n");
printf("p[18] = A04 , p[19] = A14 , p[20] = w4 , p[21] = P4 , p[22] = D4\n");
printf("p[23] = 0: Gaussian Damping , 1: Lorentz Damping\n");
return(0);
}

/* ----------------------------------------------------------------
      shared memory allocation
   ---------------------------------------------------------------- */
alloc_shm()
{
#ifdef AMIGA
   shmx = (double *) shmidx;
   shmy = (double *) shmidy;
#endif
#ifdef UNIX
   shmx = (double *) shmat(shmidx, 0,0666);
   shmy = (double *) shmat(shmidy, 0,0666);
   if((shmx < 0) || (shmy < 0)) {
      fprintf(stderr,"could not attach Shared memory, aborting !\n");
      exit(-1);
   }
#endif
return(0);
}

free_shm()
{
#ifdef UNIX
   shmdt(shmx);
   shmdt(shmy);
#endif
return(0);
}

main(argc,argv)
int argc;
char **argv;
{
int n;
double x;


   if(argc < 3) {help(); exit(0);}

   p = (double *) calloc(MAXPARM,sizeof(double));

   shmidx = atoi(argv[1]);
   shmidy = atoi(argv[2]);
   alloc_shm();
   first = atoi(argv[3]);
   last = atoi(argv[4]);
   max = atoi(argv[5]);
   for(n = 0; n < MAXPARM; n++) {
      sscanf(argv[n + 6],"%lf",&x);
      p[n] = (double) x;
   }

   lorentz_flg = (int) p[23];

   for(n = first; n <= last; n++) {
      shmy[n] = fn(shmx[n],p[0],p[1],p[2],p[3],p[4],p[5],p[6],p[7],p[8],
                           p[9],p[10],p[11],p[12],p[13],p[14],p[15],
                           p[16],p[17],p[18],p[19],p[20],p[21],p[22]);
   }
   free_shm();
   free(p);
   exit(0);
}


